@@ -1,5 +1,10 @@
+//! The empty implementation of the USDT crate.
+//!
+//! Used when the `asm` feature is disabled, or on platforms without DTrace.
+
+// Copyright 2021 Oxide Computer Company
+
 use crate::common;
-use crate::module_ident_for_provider;
 use crate::{Probe, Provider};
 use proc_macro2::TokenStream;
 use quote::quote;
@@ -15,14 +20,17 @@ pub fn compile_provider_source(
         .into_iter()
         .map(|provider| {
             let provider = Provider::from(provider);
-            let tokens = compile_provider(&provider, config);
-            let mod_name = module_ident_for_provider(&provider);
-            quote! {
-                #[macro_use]
-                pub(crate) mod #mod_name {
-                    #tokens
-                }
-            }
+            // Ensure that the name of the module in the config is set, either by the caller or
+            // defaulting to the provider name.
+            let config = crate::CompileProvidersConfig {
+                provider: Some(provider.name.clone()),
+                probe_format: config.probe_format.clone(),
+                module: match &config.module {
+                    None => Some(provider.name.clone()),
+                    other => other.clone(),
+                },
+            };
+            compile_provider(&provider, &config)
         })
         .collect::<Vec<_>>();
     Ok(quote! {
@@ -38,15 +46,14 @@ pub fn compile_provider_from_definition(
 }
 
 fn compile_provider(provider: &Provider, config: &crate::CompileProvidersConfig) -> TokenStream {
-    let mod_name = module_ident_for_provider(&provider);
     let probe_impls = provider
         .probes
         .iter()
         .map(|probe| compile_probe(&provider, probe, config))
         .collect::<Vec<_>>();
+    let module = config.module_ident();
     quote! {
-        #[macro_use]
-        pub(crate) mod #mod_name {
+        mod #module {
             #(#probe_impls)*
         }
     }
@@ -57,7 +64,7 @@ fn compile_probe(
     probe: &Probe,
     config: &crate::CompileProvidersConfig,
 ) -> TokenStream {
-    let impl_block = quote! { let _ = || ($args_lambda); };
+    let impl_block = quote! { let _ = || (__usdt_private_args_lambda()) ; };
     common::build_probe_macro(
         config,
         provider,
